from pathlib import Path
from typing import Optional, Dict, Any, List, Union

from phidata.app.phidata_app import PhidataApp, PhidataAppArgs, WorkspaceVolumeType
from phidata.utils.log import logger


class DjangoAppArgs(PhidataAppArgs):
    # -*- ECS configuration
    ecs_cluster: Optional[Any] = None
    ecs_launch_type: str = "FARGATE"
    ecs_task_cpu: str = "256"
    ecs_task_memory: str = "512"
    ecs_service_count: int = 1
    assign_public_ip: bool = True
    ecs_enable_exec: bool = True

    # -*- LoadBalancer configuration
    enable_load_balancer: bool = True
    # Change to load_balancer
    elb: Optional[Any] = None
    # Default 80 for HTTP and 443 for HTTPS
    lb_port: Optional[int] = None
    # HTTP or HTTPS
    lb_protocol: str = "HTTP"
    lb_certificate_arn: Optional[str] = None

    # -*- TargetGroup configuration
    # Default 80 for HTTP and 443 for HTTPS
    target_group_port: Optional[int] = None
    # HTTP or HTTPS
    target_group_protocol: str = "HTTP"
    health_check_protocol: Optional[str] = None
    health_check_port: Optional[str] = None
    health_check_enabled: Optional[bool] = None
    health_check_path: Optional[str] = None
    health_check_interval_seconds: Optional[int] = None
    health_check_timeout_seconds: Optional[int] = None
    healthy_threshold_count: Optional[int] = None
    unhealthy_threshold_count: Optional[int] = None

    # Add NGINX to serve static file
    enable_nginx: bool = False
    nginx_image: Optional[Any] = None
    nginx_image_name: str = "phidata/django-nginx"
    nginx_image_tag: str = "latest"
    nginx_container_port: int = 80


class DjangoApp(PhidataApp):
    def __init__(
        self,
        name: str = "django-app",
        version: str = "1",
        enabled: bool = True,
        # -*- Image Configuration,
        # Image can be provided as a DockerImage object or as image_name:image_tag
        image: Optional[Any] = None,
        image_name: str = "phidata/django-app",
        image_tag: str = "latest",
        entrypoint: Optional[str] = None,
        command: Optional[str] = "python manage.py runserver 0.0.0.0:8000",
        # Install python dependencies using a requirements.txt file,
        install_requirements: bool = False,
        # Path to the requirements.txt file relative to the workspace_root,
        requirements_file: str = "requirements.txt",
        # -*- Debug Mode
        debug_mode: bool = False,
        # Overwrite the PYTHONPATH env var, default: False
        set_python_path: bool = False,
        # Set the python_path, default: workspace_volume_container_path,
        python_path: Optional[str] = None,
        # Add to the PYTHONPATH env var. If python_path is set, this is ignored,
        add_python_path: Optional[str] = None,
        # Add env variables to container env,
        env: Optional[Dict[str, Any]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Add secret variables to container env,
        secrets: Optional[Dict[str, Any]] = None,
        # Read secret variables from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Read secret variables from AWS Secrets,
        aws_secrets: Optional[Any] = None,
        # -*- Container Configuration,
        container_name: Optional[str] = None,
        # Open a container port if open_container_port=True,
        open_container_port: bool = True,
        # Port number on the container,
        container_port: int = 8000,
        # Port name: Only used by the K8sContainer,
        container_port_name: str = "http",
        # Host port: Only used by the DockerContainer,
        container_host_port: int = 8000,
        # Mount the workspace directory on the container,
        mount_workspace: bool = False,
        workspace_volume_name: Optional[str] = None,
        workspace_volume_type: Optional[WorkspaceVolumeType] = None,
        # Path to mount the workspace volume inside the container,
        workspace_volume_container_path: str = "/usr/local/app",
        # Mount the workspace from the host machine,
        # If None, use the workspace_root_path,
        workspace_volume_host_path: Optional[str] = None,
        # Add labels to the container,
        container_labels: Optional[Dict[str, Any]] = None,
        # Run container in the background and return a Container object,
        container_detach: bool = True,
        # Enable auto-removal of the container on daemon side when the container’s process exits,
        container_auto_remove: bool = True,
        # Remove the container when it has finished running. Default: True,
        container_remove: bool = True,
        # Username or UID to run commands as inside the container,
        container_user: Optional[Union[str, int]] = None,
        # Keep STDIN open even if not attached.,
        container_stdin_open: bool = True,
        # Return logs from STDOUT when container_detach=False,
        container_stdout: Optional[bool] = True,
        # Return logs from STDERR when container_detach=False,
        container_stderr: Optional[bool] = True,
        container_tty: bool = True,
        # Specify a test to perform to check that the container is healthy.,
        container_healthcheck: Optional[Dict[str, Any]] = None,
        # Optional hostname for the container.,
        container_hostname: Optional[str] = None,
        # Platform in the format os[/arch[/variant]].,
        container_platform: Optional[str] = None,
        # Path to the working directory.,
        container_working_dir: Optional[str] = None,
        # Restart the container when it exits. Configured as a dictionary with keys:,
        # Name: One of on-failure, or always.,
        # MaximumRetryCount: Number of times to restart the container on failure.,
        # For example: {"Name": "on-failure", "MaximumRetryCount": 5},
        container_restart_policy_docker: Optional[Dict[str, Any]] = None,
        # Add volumes to DockerContainer,
        # container_volumes is a dictionary which adds the volumes to mount,
        # inside the container. The key is either the host path or a volume name,,
        # and the value is a dictionary with 2 keys:,
        #   bind - The path to mount the volume inside the container,
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.,
        # For example:,
        # {,
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},,
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'},
        # },
        container_volumes_docker: Optional[Dict[str, dict]] = None,
        # Add ports to DockerContainer,
        # The keys of the dictionary are the ports to bind inside the container,,
        # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.,
        # The values of the dictionary are the corresponding ports to open on the host, which can be either:,
        #   - The port number, as an integer.,
        #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.,
        #   - None, to assign a random host port. For example, {'2222/tcp': None}.,
        #   - A tuple of (address, port) if you want to specify the host interface.,
        #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.,
        #   - A list of integers, if you want to bind multiple host ports to a single container port.,
        #       For example, {'1111/tcp': [1234, 4567]}.,
        container_ports_docker: Optional[Dict[str, Any]] = None,
        # -*- ECS configuration,
        ecs_cluster: Optional[Any] = None,
        ecs_launch_type: str = "FARGATE",
        ecs_task_cpu: str = "256",
        ecs_task_memory: str = "512",
        ecs_service_count: int = 1,
        assign_public_ip: bool = True,
        ecs_enable_exec: bool = True,
        # -*- LoadBalancer configuration,
        enable_load_balancer: bool = True,
        # Change to load_balancer,
        elb: Optional[Any] = None,
        # Default 80 for HTTP and 443 for HTTPS,
        lb_port: Optional[int] = None,
        # HTTP or HTTPS,
        lb_protocol: str = "HTTP",
        lb_certificate_arn: Optional[str] = None,
        # -*- TargetGroup configuration,
        # Default 80 for HTTP and 443 for HTTPS,
        target_group_port: Optional[str] = None,
        # HTTP or HTTPS,
        target_group_protocol: str = "HTTP",
        health_check_protocol: Optional[str] = None,
        health_check_port: Optional[str] = None,
        health_check_enabled: Optional[bool] = None,
        health_check_path: Optional[str] = None,
        health_check_interval_seconds: Optional[int] = None,
        health_check_timeout_seconds: Optional[int] = None,
        healthy_threshold_count: Optional[int] = None,
        unhealthy_threshold_count: Optional[int] = None,
        # Add NGINX to serve static file
        enable_nginx: bool = False,
        nginx_image: Optional[Any] = None,
        nginx_image_name: str = "phidata/django-nginx",
        nginx_image_tag: str = "latest",
        # -*- AWS configuration
        aws_subnets: Optional[List[str]] = None,
        aws_security_groups: Optional[List[Any]] = None,
        # Other args,
        print_env_on_load: bool = False,
        skip_create: bool = False,
        skip_read: bool = False,
        skip_update: bool = False,
        recreate_on_update: bool = False,
        skip_delete: bool = False,
        wait_for_create: bool = True,
        wait_for_update: bool = True,
        wait_for_delete: bool = True,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 50,
        # If True, skip resource creation if active resources with the same name exist.,
        use_cache: bool = True,
        **kwargs,
    ):
        super().__init__()
        try:
            self.args: DjangoAppArgs = DjangoAppArgs(
                name=name,
                version=version,
                enabled=enabled,
                image=image,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                install_requirements=install_requirements,
                requirements_file=requirements_file,
                debug_mode=debug_mode,
                set_python_path=set_python_path,
                python_path=python_path,
                add_python_path=add_python_path,
                env=env,
                env_file=env_file,
                secrets=secrets,
                secrets_file=secrets_file,
                aws_secrets=aws_secrets,
                container_name=container_name,
                open_container_port=open_container_port,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_volume_type=workspace_volume_type,
                workspace_volume_container_path=workspace_volume_container_path,
                workspace_volume_host_path=workspace_volume_host_path,
                container_labels=container_labels,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                container_user=container_user,
                container_stdin_open=container_stdin_open,
                container_stdout=container_stdout,
                container_stderr=container_stderr,
                container_tty=container_tty,
                container_healthcheck=container_healthcheck,
                container_hostname=container_hostname,
                container_platform=container_platform,
                container_working_dir=container_working_dir,
                container_restart_policy_docker=container_restart_policy_docker,
                container_volumes_docker=container_volumes_docker,
                container_ports_docker=container_ports_docker,
                ecs_cluster=ecs_cluster,
                ecs_launch_type=ecs_launch_type,
                ecs_task_cpu=ecs_task_cpu,
                ecs_task_memory=ecs_task_memory,
                ecs_service_count=ecs_service_count,
                assign_public_ip=assign_public_ip,
                ecs_enable_exec=ecs_enable_exec,
                enable_load_balancer=enable_load_balancer,
                elb=elb,
                lb_port=lb_port,
                lb_protocol=lb_protocol,
                lb_certificate_arn=lb_certificate_arn,
                target_group_port=target_group_port,
                target_group_protocol=target_group_protocol,
                health_check_protocol=health_check_protocol,
                health_check_port=health_check_port,
                health_check_enabled=health_check_enabled,
                health_check_path=health_check_path,
                health_check_interval_seconds=health_check_interval_seconds,
                health_check_timeout_seconds=health_check_timeout_seconds,
                healthy_threshold_count=healthy_threshold_count,
                unhealthy_threshold_count=unhealthy_threshold_count,
                enable_nginx=enable_nginx,
                nginx_image=nginx_image,
                nginx_image_name=nginx_image_name,
                nginx_image_tag=nginx_image_tag,
                aws_subnets=aws_subnets,
                aws_security_groups=aws_security_groups,
                print_env_on_load=print_env_on_load,
                skip_create=skip_create,
                skip_read=skip_read,
                skip_update=skip_update,
                recreate_on_update=recreate_on_update,
                skip_delete=skip_delete,
                wait_for_creation=wait_for_create,
                wait_for_update=wait_for_update,
                wait_for_deletion=wait_for_delete,
                waiter_delay=waiter_delay,
                waiter_max_attempts=waiter_max_attempts,
                use_cache=use_cache,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid")
            raise

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(self, docker_build_context: Any) -> Optional[Any]:
        from phidata.docker.resource.group import (
            DockerNetwork,
            DockerContainer,
            DockerResourceGroup,
            DockerBuildContext,
        )
        from phidata.types.context import ContainerPathContext

        app_name = self.args.name

        if self.workspace_root_path is None:
            raise Exception("Invalid workspace_root_path")
        workspace_name = self.workspace_root_path.stem

        logger.debug(f"Building DockerResourceGroup: {app_name} for {workspace_name}")

        if docker_build_context is None or not isinstance(
            docker_build_context, DockerBuildContext
        ):
            raise Exception(f"Invalid DockerBuildContext: {type(docker_build_context)}")

        container_paths: Optional[ContainerPathContext] = self.get_container_paths(
            add_ws_name_to_ws_root=False
        )
        if container_paths is None:
            raise Exception("Invalid ContainerPathContext")
        logger.debug(f"ContainerPaths: {container_paths.json(indent=2)}")

        # Get Container Environment
        container_env: Dict[str, str] = self.get_docker_container_env(
            container_paths=container_paths
        )

        # Get Container Volumes
        container_volumes = self.get_docker_container_volumes(
            container_paths=container_paths
        )

        # Get Container Ports
        container_ports: Dict[str, int] = self.get_docker_container_ports()

        # -*- Create Docker Container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=self.get_image_str(),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove
            if not self.args.debug_mode
            else False,
            remove=self.args.container_remove if not self.args.debug_mode else False,
            healthcheck=self.args.container_healthcheck,
            hostname=self.args.container_hostname,
            labels=self.args.container_labels,
            environment=container_env,
            network=docker_build_context.network,
            platform=self.args.container_platform,
            ports=container_ports if len(container_ports) > 0 else None,
            restart_policy=self.get_container_restart_policy_docker(),
            stdin_open=self.args.container_stdin_open,
            stderr=self.args.container_stderr,
            stdout=self.args.container_stdout,
            tty=self.args.container_tty,
            user=self.args.container_user,
            volumes=container_volumes if len(container_volumes) > 0 else None,
            working_dir=self.args.container_working_dir,
            use_cache=self.args.use_cache,
        )

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
            images=[self.args.image] if self.args.image else None,
        )
        return docker_rg

    def init_docker_resource_groups(self, docker_build_context: Any) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            from collections import OrderedDict

            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## Aws Resources
    ######################################################

    def get_aws_rg(self, aws_build_context: Any) -> Optional[Any]:
        from phidata.docker.resource.image import DockerImage
        from phidata.aws.resource.group import (
            AwsResourceGroup,
            EcsCluster,
            EcsContainer,
            EcsTaskDefinition,
            EcsService,
            EcsVolume,
            LoadBalancer,
            TargetGroup,
            Listener,
            Subnet,
        )
        from phidata.types.context import ContainerPathContext
        from phidata.utils.common import get_default_volume_name

        app_name = self.name

        if self.workspace_root_path is None:
            raise Exception("Invalid workspace_root_path")
        workspace_name = self.workspace_root_path.stem

        logger.debug(f"Building AwsResourceGroup: {app_name} for {workspace_name}")

        container_paths: Optional[ContainerPathContext] = self.get_container_paths(
            add_ws_name_to_ws_root=False
        )
        if container_paths is None:
            raise Exception("Invalid ContainerPathContext")
        logger.debug(f"ContainerPaths: {container_paths.json(indent=2)}")

        # Get Container Environment
        container_env: Dict[str, str] = self.get_ecs_container_env(
            container_paths=container_paths
        )

        # -*- Create ECS cluster
        ecs_cluster = self.args.ecs_cluster
        if ecs_cluster is None:
            ecs_cluster = EcsCluster(
                name=f"{app_name}-cluster",
                ecs_cluster_name=app_name,
                capacity_providers=[self.args.ecs_launch_type],
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )

        # -*- Create Load Balancer
        load_balancer = self.args.elb
        if load_balancer is None:
            load_balancer = LoadBalancer(
                name=f"{app_name}-lb",
                subnets=self.args.aws_subnets,
                security_groups=self.args.aws_security_groups,
                protocol=self.args.lb_protocol,
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )

        # Get VPC ID from subnets
        vpc_ids = set()
        vpc_id = None
        if self.args.aws_subnets is not None:
            for subnet in self.args.aws_subnets:
                _vpc = Subnet(id=subnet).get_vpc_id()
                vpc_ids.add(_vpc)
                if len(vpc_ids) != 1:
                    raise ValueError("Subnets must be in the same VPC")
            vpc_id = vpc_ids.pop() if len(vpc_ids) == 1 else None

        # -*- Create Target Group
        tg_port = self.args.target_group_port
        if tg_port is None:
            tg_port = (
                self.args.nginx_container_port
                if self.args.enable_nginx
                else self.get_container_port()
            )
        target_group = TargetGroup(
            name=f"{app_name}-tg",
            port=tg_port,
            protocol=self.args.target_group_protocol,
            vpc_id=vpc_id,
            health_check_protocol=self.args.health_check_protocol,
            health_check_port=self.args.health_check_port,
            health_check_enabled=self.args.health_check_enabled,
            health_check_path=self.args.health_check_path,
            health_check_interval_seconds=self.args.health_check_interval_seconds,
            health_check_timeout_seconds=self.args.health_check_timeout_seconds,
            healthy_threshold_count=self.args.healthy_threshold_count,
            unhealthy_threshold_count=self.args.unhealthy_threshold_count,
            target_type="ip",
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )

        # -*- Create Listener
        lb_port = self.args.lb_port
        if lb_port is None:
            lb_port = 443 if self.args.lb_protocol == "HTTPS" else 80
        listener = Listener(
            name=f"{app_name}-listener",
            protocol=self.args.lb_protocol,
            port=lb_port,
            load_balancer=load_balancer,
            target_group=target_group,
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )
        if self.args.lb_certificate_arn is not None:
            listener.certificates = [{"CertificateArn": self.args.lb_certificate_arn}]

        # -*- Create ECS Containers
        container_command: List = []
        if isinstance(self.args.command, list):
            container_command = self.args.command
        elif isinstance(self.args.command, str):
            container_command = self.args.command.strip().split()
        else:
            logger.error(f"Invalid command: {self.args.command}")
        django_container = EcsContainer(
            name=app_name,
            image=self.get_image_str(),
            essential=True,
            port_mappings=[{"containerPort": self.get_container_port()}],
            command=container_command,
            environment=[{"name": k, "value": v} for k, v in container_env.items()],
            log_configuration={
                "logDriver": "awslogs",
                "options": {
                    "awslogs-group": app_name,
                    "awslogs-region": self.aws_region,
                    "awslogs-create-group": "true",
                    "awslogs-stream-prefix": app_name,
                },
            },
            linux_parameters={"initProcessEnabled": True},
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )
        nginx_container = None
        nginx_shared_volume = None
        if self.args.enable_nginx:
            nginx_container_name = f"{app_name}-nginx"
            nginx_shared_volume = EcsVolume(name=get_default_volume_name(app_name))
            nginx_image_str = (
                f"{self.args.nginx_image_name}:{self.args.nginx_image_tag}"
            )
            if self.args.nginx_image and isinstance(self.args.nginx_image, DockerImage):
                nginx_image_str = self.args.nginx_image.get_image_str()
            nginx_container = EcsContainer(
                name=nginx_container_name,
                image=nginx_image_str,
                essential=True,
                port_mappings=[{"containerPort": self.args.nginx_container_port}],
                environment=[{"name": k, "value": v} for k, v in container_env.items()],
                log_configuration={
                    "logDriver": "awslogs",
                    "options": {
                        "awslogs-group": app_name,
                        "awslogs-region": self.aws_region,
                        "awslogs-create-group": "true",
                        "awslogs-stream-prefix": nginx_container_name,
                    },
                },
                mount_points=[
                    {
                        "sourceVolume": nginx_shared_volume.name,
                        "containerPath": container_paths.workspace_root,
                    }
                ],
                linux_parameters={"initProcessEnabled": True},
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
                wait_for_creation=self.args.wait_for_creation,
                wait_for_deletion=self.args.wait_for_deletion,
            )
            if django_container:
                django_container.mount_points = [
                    {
                        "sourceVolume": nginx_shared_volume.name,
                        "containerPath": container_paths.workspace_root,
                    }
                ]
            else:
                logger.error("Could not add volume to django_container")

        # -*- Create ECS Task Definition
        ecs_task_definition = EcsTaskDefinition(
            name=f"{app_name}-td",
            family=app_name,
            network_mode="awsvpc",
            cpu=self.args.ecs_task_cpu,
            memory=self.args.ecs_task_memory,
            containers=[django_container],
            requires_compatibilities=[self.args.ecs_launch_type],
            add_ecs_exec_policy=self.args.ecs_enable_exec,
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )
        if self.args.enable_nginx:
            if nginx_container:
                if ecs_task_definition.containers:
                    ecs_task_definition.containers.append(nginx_container)
                else:
                    logger.error("ecs_task_definition.containers None")
            else:
                logger.error("nginx_container None")
            if nginx_shared_volume:
                ecs_task_definition.volumes = [nginx_shared_volume]

        aws_vpc_config: Dict[str, Any] = {}
        if self.args.aws_subnets is not None:
            aws_vpc_config["subnets"] = self.args.aws_subnets
        if self.args.aws_security_groups is not None:
            aws_vpc_config["securityGroups"] = self.args.aws_security_groups
        if self.args.assign_public_ip:
            aws_vpc_config["assignPublicIp"] = "ENABLED"

        # -*- Create ECS Service
        ecs_service = EcsService(
            name=f"{app_name}-service",
            desired_count=self.args.ecs_service_count,
            launch_type=self.args.ecs_launch_type,
            cluster=ecs_cluster,
            task_definition=ecs_task_definition,
            target_group=target_group,
            target_container_name=nginx_container.name
            if self.args.enable_nginx and nginx_container
            else django_container.name,
            target_container_port=self.args.nginx_container_port
            if self.args.enable_nginx
            else self.get_container_port(),
            network_configuration={"awsvpcConfiguration": aws_vpc_config},
            # Force delete the service.
            force_delete=True,
            # Force a new deployment of the service on update.
            force_new_deployment=True,
            enable_execute_command=self.args.ecs_enable_exec,
            skip_create=self.args.skip_create,
            skip_delete=self.args.skip_delete,
            wait_for_creation=self.args.wait_for_creation,
            wait_for_deletion=self.args.wait_for_deletion,
        )

        # -*- Create AwsResourceGroup
        return AwsResourceGroup(
            name=app_name,
            enabled=self.enabled,
            ecs_clusters=[ecs_cluster],
            ecs_task_definitions=[ecs_task_definition],
            ecs_services=[ecs_service],
            load_balancers=[load_balancer],
            target_groups=[target_group],
            listeners=[listener],
        )

    def init_aws_resource_groups(self, aws_build_context: Any) -> None:
        aws_rg = self.get_aws_rg(aws_build_context)
        if aws_rg is not None:
            from collections import OrderedDict

            if self.aws_resource_groups is None:
                self.aws_resource_groups = OrderedDict()
            self.aws_resource_groups[aws_rg.name] = aws_rg



from typing import Optional, List, Union

from phidata.app.base_app import BaseApp
from phidata.app.phidata_app import PhidataApp
from phidata.app.group import AppGroup, get_apps_from_app_groups
from phidata.aws.args import AwsArgs
from phidata.aws.manager import AwsManager
from phidata.aws.resource.group import AwsResourceGroup
from phidata.infra.config import InfraConfig
from phidata.utils.log import logger


class AwsConfig(InfraConfig):
    def __init__(
        self,
        env: Optional[str] = "prd",
        version: Optional[str] = None,
        enabled: bool = True,
        apps: Optional[List[Union[BaseApp, PhidataApp]]] = None,
        app_groups: Optional[List[AppGroup]] = None,
        # AwsResourceGroups to deploy
        resources: Optional[Union[AwsResourceGroup, List[AwsResourceGroup]]] = None,
        # Resources dir where aws manifests are stored
        resources_dir: str = "aws",
        # Aws params for this Config
        # Override the aws params from WorkspaceConfig if provided
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
    ):
        super().__init__()
        _resources: Optional[List[AwsResourceGroup]] = None
        if resources is not None:
            if isinstance(resources, list):
                for _r in resources:
                    if isinstance(_r, AwsResourceGroup):
                        if _resources is None:
                            _resources = []
                        _resources.append(_r)
                    else:
                        logger.error(f"Invalid resource group: {_r}")
            elif isinstance(resources, AwsResourceGroup):
                _resources = [resources]

        _apps = apps if apps is not None else []
        if app_groups is not None:
            _apps.extend(get_apps_from_app_groups(app_groups))

        try:
            self.args: AwsArgs = AwsArgs(
                env=env,
                version=version,
                enabled=enabled,
                apps=_apps,
                resources=_resources,
                resources_dir=resources_dir,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def resources(self) -> Optional[List[AwsResourceGroup]]:
        return self.args.resources if self.args else None

    @property
    def resources_dir(self) -> Optional[str]:
        return self.args.resources_dir if self.args else None

    @resources_dir.setter
    def resources_dir(self, resources_dir: str) -> None:
        if self.args is not None and resources_dir is not None:
            self.args.resources_dir = resources_dir

    def is_valid(self) -> bool:
        return True

    def get_aws_manager(self) -> Optional[AwsManager]:
        return AwsManager(aws_args=self.args)



from typing import Optional, Dict, Any, Union, List
from collections import OrderedDict

from phidata.app.base_app import BaseApp, BaseAppArgs, WorkspaceVolumeType
from phidata.types.context import ContainerPathContext
from phidata.utils.log import logger


class DockerAppArgs(BaseAppArgs):
    # -*- Resources Volume
    # Mount a resources directory on the container
    mount_resources: bool = False
    # Resources directory relative to the workspace_root
    resources_dir: str = "workspace/resources"
    # Path to mount the resources_dir
    resources_dir_container_path: str = "/mnt/resources"

    # -*- Container Configuration
    container_name: Optional[str] = None
    container_labels: Optional[Dict[str, str]] = None
    # Run container in the background and return a Container object
    container_detach: bool = True
    # Enable auto-removal of the container on daemon side when the container’s process exits
    container_auto_remove: bool = True
    # Remove the container when it has finished running. Default: True
    container_remove: bool = True
    # Username or UID to run commands as inside the container
    container_user: Optional[Union[str, int]] = None
    # Keep STDIN open even if not attached
    container_stdin_open: bool = True
    # Return logs from STDOUT when container_detach=False
    container_stdout: Optional[bool] = True
    # Return logs from STDERR when container_detach=False
    container_stderr: Optional[bool] = True
    container_tty: bool = True
    # Specify a test to perform to check that the container is healthy
    container_healthcheck: Optional[Dict[str, Any]] = None
    # Optional hostname for the container
    container_hostname: Optional[str] = None
    # Platform in the format os[/arch[/variant]]
    container_platform: Optional[str] = None
    # Path to the working directory
    container_working_dir: Optional[str] = None
    # Restart the container when it exits. Configured as a dictionary with keys:
    # Name: One of on-failure, or always.
    # MaximumRetryCount: Number of times to restart the container on failure.
    # For example: {"Name": "on-failure", "MaximumRetryCount": 5}
    container_restart_policy: Optional[Dict[str, Any]] = None
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which adds the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes: Optional[Dict[str, dict]] = None
    # Add ports to DockerContainer
    # The keys of the dictionary are the ports to bind inside the container,
    # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.
    # The values of the dictionary are the corresponding ports to open on the host, which can be either:
    #   - The port number, as an integer.
    #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
    #   - None, to assign a random host port. For example, {'2222/tcp': None}.
    #   - A tuple of (address, port) if you want to specify the host interface.
    #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.
    #   - A list of integers, if you want to bind multiple host ports to a single container port.
    #       For example, {'1111/tcp': [1234, 4567]}.
    container_ports: Optional[Dict[str, Any]] = None


class DockerApp(BaseApp):
    def __init__(self) -> None:
        super().__init__()

        # Args for the AwsAppArgs, updated by the subclass
        self.args: DockerAppArgs = DockerAppArgs()

        # Dict of DockerResourceGroups
        # Type: Optional[Dict[str, DockerResourceGroup]]
        self.docker_resource_groups: Optional[Dict[str, Any]] = None

    @property
    def container_name(self) -> str:
        from phidata.utils.common import get_default_container_name

        return (
            self.args.container_name
            if self.args.container_name
            else get_default_container_name(self.name)
        )

    @container_name.setter
    def container_name(self, container_name: str) -> None:
        if self.args is not None and container_name is not None:
            self.args.container_name = container_name

    @property
    def container_restart_policy(self) -> Optional[Dict[str, Any]]:
        return self.args.container_restart_policy

    @container_restart_policy.setter
    def container_restart_policy(
        self, container_restart_policy: Dict[str, Any]
    ) -> None:
        if self.args is not None and container_restart_policy is not None:
            self.args.container_restart_policy = container_restart_policy

    def get_container_env_docker(
        self, container_paths: ContainerPathContext
    ) -> Dict[str, str]:
        from phidata.constants import (
            PYTHONPATH_ENV_VAR,
            PHIDATA_RUNTIME_ENV_VAR,
            SCRIPTS_DIR_ENV_VAR,
            STORAGE_DIR_ENV_VAR,
            META_DIR_ENV_VAR,
            PRODUCTS_DIR_ENV_VAR,
            NOTEBOOKS_DIR_ENV_VAR,
            WORKFLOWS_DIR_ENV_VAR,
            WORKSPACE_ROOT_ENV_VAR,
            WORKSPACE_CONFIG_DIR_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}
        container_env.update(
            {
                PHIDATA_RUNTIME_ENV_VAR: "docker",
                SCRIPTS_DIR_ENV_VAR: container_paths.scripts_dir or "",
                STORAGE_DIR_ENV_VAR: container_paths.storage_dir or "",
                META_DIR_ENV_VAR: container_paths.meta_dir or "",
                PRODUCTS_DIR_ENV_VAR: container_paths.products_dir or "",
                NOTEBOOKS_DIR_ENV_VAR: container_paths.notebooks_dir or "",
                WORKFLOWS_DIR_ENV_VAR: container_paths.workflows_dir or "",
                WORKSPACE_ROOT_ENV_VAR: container_paths.workspace_root or "",
                WORKSPACE_CONFIG_DIR_ENV_VAR: container_paths.workspace_config_dir
                or "",
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": container_paths.requirements_file or "",
                "MOUNT_WORKSPACE": str(self.args.mount_workspace),
                "WORKSPACE_DIR_CONTAINER_PATH": str(
                    self.args.workspace_dir_container_path
                ),
                "MOUNT_RESOURCES": str(self.args.mount_resources),
                "RESOURCES_DIR_CONTAINER_PATH": str(
                    self.args.resources_dir_container_path
                ),
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            }
        )

        if self.args.set_python_path:
            python_path = self.args.python_path
            if python_path is None:
                python_path = container_paths.workspace_root
                if (
                    self.args.mount_resources
                    and self.args.resources_dir_container_path is not None
                ):
                    python_path = "{}:{}".format(
                        python_path, self.args.resources_dir_container_path
                    )
                if self.args.add_python_paths is not None:
                    python_path = "{}:{}".format(
                        python_path, ":".join(self.args.add_python_paths)
                    )
            if python_path is not None:
                container_env[PYTHONPATH_ENV_VAR] = python_path

        # Set aws env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(
                {k: str(v) for k, v in env_data_from_file.items() if v is not None}
            )

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data()
        if secret_data_from_file is not None:
            container_env.update(
                {k: str(v) for k, v in secret_data_from_file.items() if v is not None}
            )

        # Update the container env with user provided env
        # this overwrites any existing variables with the same key
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(
                {k: str(v) for k, v in self.args.env.items() if v is not None}
            )

        # logger.debug("Container Environment: {}".format(container_env))
        return container_env

    def get_container_volumes_docker(
        self, container_paths: ContainerPathContext
    ) -> Dict[str, dict]:
        from phidata.utils.common import get_default_volume_name

        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return {}

        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = self.args.container_volumes or {}

        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_volume_container_path_str = container_paths.workspace_root
            if (
                self.args.workspace_volume_type is None
                or self.args.workspace_volume_type == WorkspaceVolumeType.HostPath
            ):
                workspace_volume_host_path = self.args.workspace_dir or str(
                    self.workspace_root_path
                )
                logger.debug(f"Mounting: {workspace_volume_host_path}")
                logger.debug(f"\tto: {workspace_volume_container_path_str}")
                container_volumes[workspace_volume_host_path] = {
                    "bind": workspace_volume_container_path_str,
                    "mode": "rw",
                }
            elif self.args.workspace_volume_type == WorkspaceVolumeType.EmptyDir:
                workspace_volume_name = self.args.workspace_volume_name
                if workspace_volume_name is None:
                    workspace_volume_name = get_default_volume_name(
                        container_paths.workspace_name or "ws"
                    )
                logger.debug(f"Mounting: {workspace_volume_name}")
                logger.debug(f"\tto: {workspace_volume_container_path_str}")
                container_volumes[workspace_volume_name] = {
                    "bind": workspace_volume_container_path_str,
                    "mode": "rw",
                }
            else:
                logger.error(f"{self.args.workspace_volume_type.value} not supported")

        # Create a volume for the resources dir
        if self.args.mount_resources:
            resources_dir_path = str(
                self.workspace_root_path.joinpath(self.args.resources_dir)
            )
            logger.debug(f"Mounting: {resources_dir_path}")
            logger.debug(f"\tto: {self.args.resources_dir_container_path}")
            container_volumes[resources_dir_path] = {
                "bind": self.args.resources_dir_container_path,
                "mode": "ro",
            }

        return container_volumes

    def get_container_ports_docker(self) -> Dict[str, int]:
        # container_ports is a dictionary which configures the ports to bind
        # inside the container. The key is the port to bind inside the container
        #   either as an integer or a string in the form port/protocol
        # and the value is the corresponding port to open on the host.
        # For example:
        #   {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
        container_ports: Dict[str, int] = self.args.container_ports or {}

        if self.args.open_container_port:
            container_ports[
                str(self.args.container_port)
            ] = self.args.container_host_port

        return container_ports

    def get_container_command_docker(self) -> Optional[List[str]]:
        if isinstance(self.args.command, str):
            return self.args.command.split(" ")
        return self.args.command

    def get_docker_rg(self, docker_build_context: Any) -> Optional[Any]:
        from phidata.docker.resource.group import (
            DockerNetwork,
            DockerContainer,
            DockerResourceGroup,
            DockerBuildContext,
        )

        # -*- Build Container Paths
        container_paths: Optional[ContainerPathContext] = self.get_container_paths()
        if container_paths is None:
            raise Exception("Could not build Container Paths")
        # logger.debug(f"ContainerPaths: {container_paths.json(indent=2)}")

        app_name = self.name
        workspace_name = container_paths.workspace_name
        logger.debug(f"Building DockerResourceGroup: {app_name} for {workspace_name}")

        if docker_build_context is None or not isinstance(
            docker_build_context, DockerBuildContext
        ):
            logger.error("docker_build_context must be a DockerBuildContext")
            return None

        # -*- Build Container Environment
        container_env: Dict[str, str] = self.get_container_env_docker(
            container_paths=container_paths
        )

        # -*- Build Container Volumes
        container_volumes = self.get_container_volumes_docker(
            container_paths=container_paths
        )

        # -*- Build Container Ports
        container_ports: Dict[str, int] = self.get_container_ports_docker()

        # -*- Build Container Command
        container_cmd: Optional[List[str]] = self.get_container_command_docker()
        if container_cmd:
            logger.debug("Command: {}".format(" ".join(container_cmd)))

        # -*- Create DockerContainer
        docker_container = DockerContainer(
            name=self.container_name,
            image=self.get_image_str(),
            entrypoint=self.args.entrypoint,
            command=" ".join(container_cmd) if container_cmd is not None else None,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove
            if not self.args.debug_mode
            else False,
            remove=self.args.container_remove if not self.args.debug_mode else False,
            healthcheck=self.args.container_healthcheck,
            hostname=self.args.container_hostname,
            labels=self.args.container_labels,
            environment=container_env,
            network=docker_build_context.network,
            platform=self.args.container_platform,
            ports=container_ports if len(container_ports) > 0 else None,
            restart_policy=self.container_restart_policy,
            stdin_open=self.args.container_stdin_open,
            stderr=self.args.container_stderr,
            stdout=self.args.container_stdout,
            tty=self.args.container_tty,
            user=self.args.container_user,
            volumes=container_volumes if len(container_volumes) > 0 else None,
            working_dir=self.args.container_working_dir,
            use_cache=self.args.use_cache,
        )

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
            images=[self.args.image] if self.args.image else None,
        )
        return docker_rg

    def build_docker_resource_groups(self, docker_build_context: Any) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    def get_docker_resource_groups(
        self, docker_build_context: Any, defer_api_calls: bool = False
    ) -> Optional[Dict[str, Any]]:
        if self.docker_resource_groups is None:
            self.build_docker_resource_groups(docker_build_context)
        # Comment out in production
        # if self.docker_resource_groups:
        #     logger.debug("DockerResourceGroups:")
        #     for rg_name, rg in self.docker_resource_groups.items():
        #         logger.debug(
        #             "{}: {}".format(rg_name, rg.json(exclude_none=True, indent=2))
        #         )
        return self.docker_resource_groups



from pathlib import Path
from typing import Optional, List, Dict, Union, Any

from phidata.infra.args import InfraArgs
from phidata.infra.config import InfraConfig
from phidata.workspace.settings import WorkspaceSettings
from phidata.utils.log import logger


class WorkspaceConfigArgs(InfraArgs):
    # default env for phi ws ... commands
    default_env: Optional[str] = None
    # default config for phi ws ... commands
    default_config: Optional[str] = None
    # A copy of the workspace settings
    ws_settings: Optional[WorkspaceSettings] = None

    # List of Docker configurations
    # Type: DockerConfig
    docker: Optional[List[Any]] = None
    # List of K8s configurations
    # Type: K8sConfig
    k8s: Optional[List[Any]] = None
    # List of AWS configurations
    # Type: AwsConfig
    aws: Optional[List[Any]] = None

    # Path to important directories relative to the workspace_root
    meta_dir: str
    notebooks_dir: str
    products_dir: str
    scripts_dir: str
    storage_dir: str
    workflows_dir: str
    workspace_config_dir: str


class WorkspaceConfig(InfraConfig):
    def __init__(
        self,
        version: Optional[str] = None,
        enabled: bool = True,
        # default env for phi ws ... commands
        default_env: Optional[str] = None,
        # default config for phi ws ... commands
        default_config: Optional[str] = None,
        ws_settings: Optional[WorkspaceSettings] = None,
        # List of Docker configurations
        # Type: DockerConfig
        docker: Optional[List[Any]] = None,
        # List of K8s configurations
        # Type: K8sConfig
        k8s: Optional[List[Any]] = None,
        # List of AWS configurations
        # Type: AwsConfig
        aws: Optional[List[Any]] = None,
        # Path to important directories relative to the workspace_root
        meta_dir: str = "meta",
        notebooks_dir: str = "notebooks",
        products_dir: str = "workflows",
        scripts_dir: str = "scripts",
        storage_dir: str = "storage",
        workflows_dir: str = "workflows",
        workspace_config_dir: str = "workspace",
        # -*- Environment parameters
        # NOTE: These are NOT used by the WorkspaceConfig
        # So cannot be used in the workspace files
        # Env vars added to local env when running workflows
        local_env: Optional[Dict[str, str]] = None,
        # Load local env using env file
        local_env_file: Optional[Union[str, Path]] = None,
        # Env vars added to docker env when running workflows
        docker_env: Optional[Dict[str, str]] = None,
        # Load docker env using env file
        docker_env_file: Optional[Union[str, Path]] = None,
        # Env vars added to k8s env when running workflows
        k8s_env: Optional[Dict[str, str]] = None,
        # Load k8s env using env file
        k8s_env_file: Optional[Union[str, Path]] = None,
        # -*- AWS parameters
        # Common aws params used by apps, resources and data assets
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
        # -*- `phi` cli parameters
        # Set to True if `phi` should continue creating
        # resources after a resource creation has failed
        continue_on_create_failure: bool = False,
        # Set to True if `phi` should continue deleting
        # resources after a resource deleting has failed
        # Defaults to True because we normally want to continue deleting
        continue_on_delete_failure: bool = True,
        # Set to True if `phi` should continue patching
        # resources after a resource patch has failed
        continue_on_patch_failure: bool = False,
    ):
        super().__init__()

        try:
            self.args: WorkspaceConfigArgs = WorkspaceConfigArgs(
                version=version,
                enabled=enabled,
                default_env=default_env,
                default_config=default_config,
                ws_settings=ws_settings,
                docker=docker,
                k8s=k8s,
                aws=aws,
                meta_dir=meta_dir,
                notebooks_dir=notebooks_dir,
                products_dir=products_dir,
                scripts_dir=scripts_dir,
                storage_dir=storage_dir,
                workflows_dir=workflows_dir,
                workspace_config_dir=workspace_config_dir,
                local_env=local_env,
                local_env_file=local_env_file,
                docker_env=docker_env,
                docker_env_file=docker_env_file,
                k8s_env=k8s_env,
                k8s_env_file=k8s_env_file,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
                continue_on_create_failure=continue_on_create_failure,
                continue_on_delete_failure=continue_on_delete_failure,
                continue_on_patch_failure=continue_on_patch_failure,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def default_env(self) -> Optional[str]:
        return self.args.default_env if self.args else None

    @default_env.setter
    def default_env(self, default_env: str) -> None:
        if self.args is not None and default_env is not None:
            self.args.default_env = default_env

    @property
    def default_config(self) -> Optional[str]:
        return self.args.default_config if self.args else None

    @default_config.setter
    def default_config(self, default_config: str) -> None:
        if self.args is not None and default_config is not None:
            self.args.default_config = default_config

    @property
    def ws_settings(self) -> Optional[WorkspaceSettings]:
        return self.args.ws_settings if self.args else None

    @ws_settings.setter
    def ws_settings(self, ws_settings: WorkspaceSettings) -> None:
        if self.args is not None and ws_settings is not None:
            self.args.ws_settings = ws_settings

    @property
    def docker(self) -> Optional[List[Any]]:
        return self.args.docker if self.args else None

    @docker.setter
    def docker(self, docker: Any) -> None:
        if self.args is not None and docker is not None:
            self.args.docker = docker

    @property
    def k8s(self) -> Optional[List[Any]]:
        return self.args.k8s if self.args else None

    @k8s.setter
    def k8s(self, k8s: Any) -> None:
        if self.args is not None and k8s is not None:
            self.args.k8s = k8s

    @property
    def aws(self) -> Optional[List[Any]]:
        return self.args.aws if self.args else None

    @aws.setter
    def aws(self, aws: Any) -> None:
        if self.args is not None and aws is not None:
            self.args.aws = aws

    def is_valid(self) -> bool:
        if self.docker is not None:
            from phidata.docker.config import DockerConfig

            if not isinstance(self.docker, list):
                raise ValueError("docker should be a list")
            for dc in self.docker:
                if not isinstance(dc, DockerConfig):
                    raise ValueError(f"Invalid DockerConfig: {dc}")

        if self.k8s is not None:
            from phidata.k8s.config import K8sConfig

            if not isinstance(self.k8s, list):
                raise ValueError("k8s should be a list")
            for kc in self.k8s:
                if not isinstance(kc, K8sConfig):
                    raise ValueError(f"Invalid K8sConfig: {kc}")

        if self.aws is not None:
            from phidata.aws.config import AwsConfig

            if not isinstance(self.aws, list):
                raise ValueError("aws should be a list")
            for awsc in self.aws:
                if not isinstance(awsc, AwsConfig):
                    raise ValueError(f"Invalid AwsConfig: {awsc}")
        return True

    @classmethod
    def from_file(cls, path: Path) -> "WorkspaceConfig":
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist")
        if not path.is_file():
            raise ValueError(f"Path {path} is not a file")
        if not path.suffix == ".py":
            raise ValueError(f"File {path} is not a python file")

        ws_config = cls()
        logger.debug(f"--^^-- Building WorkspaceConfig from: {path}")
        ws_config.workspace_root_path = path.parent.resolve()
        workspace_config_objects = {}
        try:
            from phidata.utils.get_python_objects_from_module import (
                get_python_objects_from_module,
            )

            python_objects = get_python_objects_from_module(path)
            for obj_name, obj in python_objects.items():
                _type_name = obj.__class__.__name__
                if _type_name in [
                    "WorkspaceSettings",
                    "DockerConfig",
                    "K8sConfig",
                    "AwsConfig",
                ]:
                    workspace_config_objects[obj_name] = obj
        except Exception as e:
            parent_dir = path.parent.name
            parent_parent_dir = path.parent.parent.name
            if parent_dir in ("resources", "tests") or parent_parent_dir in (
                "resources",
                "tests",
            ):
                pass
            else:
                logger.warning(f"Error in {path}: {e}")
            pass

        # logger.debug(f"workspace_config_objects: {workspace_config_objects}")
        for obj_name, obj in workspace_config_objects.items():
            _obj_type = obj.__class__.__name__
            logger.debug(f"Adding {obj_name} | Type: {_obj_type}")
            if _obj_type == "WorkspaceSettings":
                ws_config.ws_settings = obj
                try:
                    ws_config.default_env = obj.default_env or obj.dev_env
                    ws_config.default_config = obj.default_config
                except Exception as e:
                    logger.debug(f"Error loading default settings: {e}")
                    pass

                try:
                    ws_config.aws_region = obj.aws_region
                    ws_config.aws_profile = obj.aws_profile
                    ws_config.aws_config_file = obj.aws_config_file
                    ws_config.aws_shared_credentials_file = (
                        obj.aws_shared_credentials_file
                    )
                except Exception as e:
                    logger.debug(f"Error loading aws settings: {e}")
                    pass
            elif _obj_type == "DockerConfig":
                if ws_config.docker is None:
                    ws_config.docker = []
                ws_config.docker.append(obj)
            elif _obj_type == "K8sConfig":
                if ws_config.k8s is None:
                    ws_config.k8s = []
                ws_config.k8s.append(obj)
            elif _obj_type == "AwsConfig":
                if ws_config.aws is None:
                    ws_config.aws = []
                ws_config.aws.append(obj)
        return ws_config



from pathlib import Path
from typing import Optional, Dict, Union

from phidata.base import PhidataBaseArgs


class InfraArgs(PhidataBaseArgs):
    env: Optional[str] = None

    # -*- Path parameters
    # Path to the workspace root directory
    workspace_root_path: Optional[Path] = None
    # Path to the workspace config file
    workspace_config_file_path: Optional[Path] = None
    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root_path
    #   to build paths depending on the environments (local, docker, k8s)
    # defaults are set by WorkspaceConfig.__init__()
    meta_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    products_dir: Optional[str] = None
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    workflows_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None

    # -*- Environment parameters
    # NOTE: These are NOT used by the WorkspaceConfig
    # So cannot be used in the workspace files
    # Env vars added to local env when running workflows
    local_env: Optional[Dict[str, str]] = None
    # Load local env using env file
    local_env_file: Optional[Union[str, Path]] = None
    # Env vars added to docker env when running workflows
    docker_env: Optional[Dict[str, str]] = None
    # Load docker env using env file
    docker_env_file: Optional[Union[str, Path]] = None
    # Env vars added to k8s env when running workflows
    k8s_env: Optional[Dict[str, str]] = None
    # Load k8s env using env file
    k8s_env_file: Optional[Union[str, Path]] = None

    # -*- AWS parameters
    # Common aws params used by apps, resources and data assets
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None

    # -*- `phi` cli parameters
    # Set to True if `phi` should continue creating
    # resources after a resource creation has failed
    continue_on_create_failure: Optional[bool] = None
    # Set to True if `phi` should continue deleting
    # resources after a resource deleting has failed
    # Defaults to True because we normally want to continue deleting
    continue_on_delete_failure: Optional[bool] = None
    # Set to True if `phi` should continue patching
    # resources after a resource patch has failed
    continue_on_patch_failure: Optional[bool] = None



from phidata.utils.log import logger

try:
    import duckdb
except ImportError:
    logger.warning("DuckDB not available.")
    raise


def run_duckdb_query(
    duckdb_connection: duckdb.DuckDBPyConnection, sql_query: str
) -> str:
    """
    Function to run SQL queries against a duckdb database

    Args:
        duckdb_connection (duckdb.DuckDBPyConnection): duckdb connection
        sql_query (str): SQL query to run

    Returns:
        str: Result of the query
    """
    # -*- Format the SQL Query
    # Remove backticks
    formatted_sql = sql_query.replace("`", "")
    # If there are multiple statements, only run the first one
    formatted_sql = formatted_sql.split(";")[0]

    try:
        query_result = duckdb_connection.sql(formatted_sql)
        result_output = "No output"
        if query_result is not None:
            try:
                results_as_python_objects = query_result.fetchall()
                result_rows = []
                for row in results_as_python_objects:
                    if len(row) == 1:
                        result_rows.append(str(row[0]))
                    else:
                        result_rows.append(",".join(str(x) for x in row))

                result_data = "\n".join(result_rows)
                result_output = ",".join(query_result.columns) + "\n" + result_data
            except AttributeError:
                result_output = str(query_result)
        return result_output
    except duckdb.ProgrammingError as e:
        return str(e)
    except duckdb.Error as e:
        return str(e)
    except Exception as e:
        return str(e)


def describe_table_or_view(duckdb_connection: duckdb.DuckDBPyConnection, table: str):
    """
    Function to describe a table or view

    Args:
        duckdb_connection (duckdb.DuckDBPyConnection): duckdb connection
        table (str): Table or view to describe

    Returns:
        str: Description of the table or view
    """
    statement = f"select column_name, data_type from information_schema.columns where table_name='{table}';"
    table_description = run_duckdb_query(duckdb_connection, statement)
    return f"{table}\n{table_description}"



from pathlib import Path
from typing import Optional, Dict, Any, Union, List

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.types.context import ContainerPathContext
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class WorkspaceVolumeType(ExtendedEnum):
    HostPath = "HostPath"
    EmptyDir = "EmptyDir"
    AwsEbs = "AwsEbs"
    AwsEfs = "AwsEfs"
    PersistentVolume = "PersistentVolume"


class PhidataAppArgs(PhidataBaseArgs):
    name: Optional[str]

    # -*- Path parameters
    # The following args are populated by the K8sWorker and DockerWorker classes.
    # The build_resource_groups() function passes these args from the
    # WorkspaceConfig -> K8sConfig -> K8sArgs -> PhidataApp

    # Path to the workspace root directory
    workspace_root_path: Optional[Path] = None
    # Path to the workspace config file
    workspace_config_file_path: Optional[Path] = None
    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root dir
    #   to build paths depending on the environments (local, docker, k8s)
    # defaults are set by WorkspaceConfig.__init__()
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    meta_dir: Optional[str] = None
    products_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    workflows_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None

    # -*- Common environment variables from the WorkspaceConfig
    # Env vars added to docker env when building PhidataApps
    #   and running workflows
    docker_env: Optional[Dict[str, str]] = None
    # Env vars added to k8s env when building PhidataApps
    #   and running workflows
    k8s_env: Optional[Dict[str, str]] = None

    # -*- Image Configuration
    # Image can be provided as a DockerImage object or as image_name:image_tag
    image: Optional[Any] = None
    image_name: Optional[str] = None
    image_tag: Optional[str] = None
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # -*- Debug Mode
    debug_mode: bool = False

    # -*- Python Configuration
    # Install python dependencies using a requirements.txt file
    install_requirements: bool = False
    # Path to the requirements.txt file relative to the workspace_root
    requirements_file: str = "requirements.txt"
    # Set the container PYTHONPATH
    set_python_path: bool = False
    # Manually provide the PYTHONPATH. If None, defaults to the workspace_root
    python_path: Optional[str] = None
    # Add to the PYTHONPATH env var. If python_path is provided, this value is ignored
    add_python_path: Optional[str] = None

    # Container env passed to the PhidataApp
    # Add env variables to container env
    env: Optional[Dict[str, Any]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None

    # Container secrets
    # Add secret variables to container env
    secrets: Optional[Dict[str, Any]] = None
    # Read secret variables from a file in yaml format
    secrets_file: Optional[Path] = None
    # Read secret variables from AWS Secrets
    aws_secrets: Optional[Any] = None

    # -*- Container Ports
    # Open a container port if open_container_port=True
    open_container_port: bool = False
    # Port number on the container
    container_port: int = 8000
    # Port name
    container_port_name: str = "http"
    # Host port to map to the container port
    container_host_port: int = 8000

    # -*- Container Volumes
    # Mount the workspace directory on the container
    mount_workspace: bool = False
    workspace_volume_name: Optional[str] = None
    workspace_volume_type: Optional[WorkspaceVolumeType] = None
    # Path to mount the workspace volume inside the container
    workspace_volume_container_path: str = "/mnt/workspaces"
    # Mount the workspace from the host machine
    # If None, use the workspace_root_path
    workspace_volume_host_path: Optional[str] = None
    # Load the workspace from git using a git-sync sidecar container
    create_git_sync_sidecar: bool = False
    # Required to create an initial copy of the workspace
    create_git_sync_init_container: bool = True
    git_sync_image_name: str = "k8s.gcr.io/git-sync"
    git_sync_image_tag: str = "v3.1.1"
    git_sync_repo: Optional[str] = None
    git_sync_branch: Optional[str] = None
    git_sync_wait: int = 1

    # -*- Docker Configuration
    container_name: Optional[str] = None
    # Run container in the background and return a Container object.
    container_detach: bool = True
    # Enable auto-removal of the container on daemon side when the container’s process exits.
    container_auto_remove: bool = True
    # Remove the container when it has finished running. Default: True.
    container_remove: bool = True
    # Username or UID to run commands as inside the container.
    container_user: Optional[Union[str, int]] = None
    # Keep STDIN open even if not attached.
    container_stdin_open: bool = True
    # Return logs from STDOUT when container_detach=False.
    container_stdout: Optional[bool] = True
    # Return logs from STDERR when container_detach=False.
    container_stderr: Optional[bool] = True
    container_tty: bool = True
    # Specify a test to perform to check that the container is healthy.
    container_healthcheck: Optional[Dict[str, Any]] = None
    # Optional hostname for the container.
    container_hostname: Optional[str] = None
    # Platform in the format os[/arch[/variant]].
    container_platform: Optional[str] = None
    # Path to the working directory.
    container_working_dir: Optional[str] = None
    # Restart the container when it exits. Configured as a dictionary with keys:
    # Name: One of on-failure, or always.
    # MaximumRetryCount: Number of times to restart the container on failure.
    # For example: {"Name": "on-failure", "MaximumRetryCount": 5}
    container_restart_policy_docker: Optional[Dict[str, Any]] = None
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which adds the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes_docker: Optional[Dict[str, dict]] = None
    # Add ports to DockerContainer
    # The keys of the dictionary are the ports to bind inside the container,
    # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.
    # The values of the dictionary are the corresponding ports to open on the host, which can be either:
    #   - The port number, as an integer.
    #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
    #   - None, to assign a random host port. For example, {'2222/tcp': None}.
    #   - A tuple of (address, port) if you want to specify the host interface.
    #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.
    #   - A list of integers, if you want to bind multiple host ports to a single container port.
    #       For example, {'1111/tcp': [1234, 4567]}.
    container_ports_docker: Optional[Dict[str, Any]] = None
    # Add labels to the container
    container_labels: Optional[Dict[str, Any]] = None

    # -*- K8s Deployment Configuration
    replicas: int = 1
    pod_name: Optional[str] = None
    deploy_name: Optional[str] = None
    secret_name: Optional[str] = None
    configmap_name: Optional[str] = None
    # Type: ImagePullPolicy
    image_pull_policy: Optional[Any] = None
    pod_annotations: Optional[Dict[str, str]] = None
    pod_node_selector: Optional[Dict[str, str]] = None
    # Type: RestartPolicy
    deploy_restart_policy: Optional[Any] = None
    deploy_labels: Optional[Dict[str, Any]] = None
    termination_grace_period_seconds: Optional[int] = None
    # Key to spread the pods across a topology
    topology_spread_key: Optional[str] = None
    # The degree to which pods may be unevenly distributed
    topology_spread_max_skew: Optional[int] = None
    # How to deal with a pod if it doesn't satisfy the spread constraint.
    topology_spread_when_unsatisfiable: Optional[str] = None

    # -*- K8s Service Configuration
    create_service: bool = False
    service_name: Optional[str] = None
    # Type: ServiceType
    service_type: Optional[Any] = None
    # The port exposed by the service.
    service_port: int = 8000
    # The node_port exposed by the service if service_type = ServiceType.NODE_PORT
    service_node_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    service_target_port: Optional[Union[str, int]] = None
    # Extra ports exposed by the webserver service. Type: List[CreatePort]
    service_ports: Optional[List[Any]] = None
    service_labels: Optional[Dict[str, Any]] = None
    service_annotations: Optional[Dict[str, str]] = None
    # If ServiceType == ServiceType.LoadBalancer
    service_health_check_node_port: Optional[int] = None
    service_internal_traffic_policy: Optional[str] = None
    service_load_balancer_class: Optional[str] = None
    service_load_balancer_ip: Optional[str] = None
    service_load_balancer_source_ranges: Optional[List[str]] = None
    service_allocate_load_balancer_node_ports: Optional[bool] = None

    # -*- K8s Ingress Configuration
    create_ingress: bool = False
    ingress_name: Optional[str] = None
    ingress_annotations: Optional[Dict[str, str]] = None

    # -*- K8s RBAC Configuration
    use_rbac: bool = False
    # Create a Namespace with name ns_name & default values
    ns_name: Optional[str] = None
    # or Provide the full Namespace definition
    # Type: CreateNamespace
    namespace: Optional[Any] = None
    # Create a ServiceAccount with name sa_name & default values
    sa_name: Optional[str] = None
    # or Provide the full ServiceAccount definition
    # Type: CreateServiceAccount
    service_account: Optional[Any] = None
    # Create a ClusterRole with name cr_name & default values
    cr_name: Optional[str] = None
    # or Provide the full ClusterRole definition
    # Type: CreateClusterRole
    cluster_role: Optional[Any] = None
    # Create a ClusterRoleBinding with name crb_name & default values
    crb_name: Optional[str] = None
    # or Provide the full ClusterRoleBinding definition
    # Type: CreateClusterRoleBinding
    cluster_role_binding: Optional[Any] = None

    # Add additional Kubernetes resources to the App
    # Type: CreateSecret
    extra_secrets: Optional[List[Any]] = None
    # Type: CreateConfigMap
    extra_configmaps: Optional[List[Any]] = None
    # Type: CreateService
    extra_services: Optional[List[Any]] = None
    # Type: CreateDeployment
    extra_deployments: Optional[List[Any]] = None
    # Type: CreatePersistentVolume
    extra_pvs: Optional[List[Any]] = None
    # Type: CreatePVC
    extra_pvcs: Optional[List[Any]] = None
    # Type: CreateContainer
    extra_containers: Optional[List[Any]] = None
    # Type: CreateContainer
    extra_init_containers: Optional[List[Any]] = None
    # Type: CreatePort
    extra_ports: Optional[List[Any]] = None
    # Type: CreateVolume
    extra_volumes: Optional[List[Any]] = None
    # Type: CreateStorageClass
    extra_storage_classes: Optional[List[Any]] = None
    # Type: CreateCustomObject
    extra_custom_objects: Optional[List[Any]] = None
    # Type: CreateCustomResourceDefinition
    extra_crds: Optional[List[Any]] = None

    # -*- AWS configuration
    aws_subnets: Optional[List[str]] = None
    aws_security_groups: Optional[List[Any]] = None

    # Other args
    print_env_on_load: bool = False
    # If True, skip resource creation if active resources with the same name exist.
    use_cache: bool = True

    # Extra kwargs used to ensure older versions of phidata don't throw syntax errors
    extra_kwargs: Optional[Dict[str, Any]] = None

    # -*- AWS parameters
    # Common aws params used by apps, resources and data assets
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True


class PhidataApp(PhidataBase):
    """Base Class for all PhidataApps"""

    def __init__(self) -> None:
        super().__init__()

        # Cache env_data & secret_data
        self.env_data: Optional[Dict[str, Any]] = None
        self.secret_data: Optional[Dict[str, Any]] = None

        # Args for the PhidataApp, set by the subclass
        self.args: PhidataAppArgs = PhidataAppArgs()

        # Dict of DockerResourceGroups
        # Type: Optional[Dict[str, DockerResourceGroup]]
        self.docker_resource_groups: Optional[Dict[str, Any]] = None

        # Dict of KubernetesResourceGroups
        # Type: Optional[Dict[str, K8sResourceGroup]]
        self.k8s_resource_groups: Optional[Dict[str, Any]] = None

        # Dict of AwsResourceGroups
        # Type: Optional[Dict[str, AwsResourceGroup]]
        self.aws_resource_groups: Optional[Dict[str, Any]] = None

    @property
    def workspace_root_path(self) -> Optional[Path]:
        return self.args.workspace_root_path if self.args else None

    @workspace_root_path.setter
    def workspace_root_path(self, workspace_root_path: Path) -> None:
        if self.args is not None and workspace_root_path is not None:
            self.args.workspace_root_path = workspace_root_path

    @property
    def workspace_config_file_path(self) -> Optional[Path]:
        return self.args.workspace_config_file_path if self.args else None

    @workspace_config_file_path.setter
    def workspace_config_file_path(self, workspace_config_file_path: Path) -> None:
        if self.args is not None and workspace_config_file_path is not None:
            self.args.workspace_config_file_path = workspace_config_file_path

    @property
    def scripts_dir(self) -> Optional[str]:
        return self.args.scripts_dir if self.args else None

    @scripts_dir.setter
    def scripts_dir(self, scripts_dir: str) -> None:
        if self.args is not None and scripts_dir is not None:
            self.args.scripts_dir = scripts_dir

    @property
    def storage_dir(self) -> Optional[str]:
        return self.args.storage_dir if self.args else None

    @storage_dir.setter
    def storage_dir(self, storage_dir: str) -> None:
        if self.args is not None and storage_dir is not None:
            self.args.storage_dir = storage_dir

    @property
    def meta_dir(self) -> Optional[str]:
        return self.args.meta_dir if self.args else None

    @meta_dir.setter
    def meta_dir(self, meta_dir: str) -> None:
        if self.args is not None and meta_dir is not None:
            self.args.meta_dir = meta_dir

    @property
    def products_dir(self) -> Optional[str]:
        return self.args.products_dir if self.args else None

    @products_dir.setter
    def products_dir(self, products_dir: str) -> None:
        if self.args is not None and products_dir is not None:
            self.args.products_dir = products_dir

    @property
    def notebooks_dir(self) -> Optional[str]:
        return self.args.notebooks_dir if self.args else None

    @notebooks_dir.setter
    def notebooks_dir(self, notebooks_dir: str) -> None:
        if self.args is not None and notebooks_dir is not None:
            self.args.notebooks_dir = notebooks_dir

    @property
    def workflows_dir(self) -> Optional[str]:
        return self.args.workflows_dir if self.args else None

    @workflows_dir.setter
    def workflows_dir(self, workflows_dir: str) -> None:
        if self.args is not None and workflows_dir is not None:
            self.args.workflows_dir = workflows_dir

    @property
    def workspace_config_dir(self) -> Optional[str]:
        return self.args.workspace_config_dir if self.args else None

    @workspace_config_dir.setter
    def workspace_config_dir(self, workspace_config_dir: str) -> None:
        if self.args is not None and workspace_config_dir is not None:
            self.args.workspace_config_dir = workspace_config_dir

    @property
    def docker_env(self) -> Optional[Dict[str, str]]:
        return self.args.docker_env if self.args else None

    @docker_env.setter
    def docker_env(self, docker_env: Dict[str, str]) -> None:
        if self.args is not None and docker_env is not None:
            self.args.docker_env = docker_env

    @property
    def k8s_env(self) -> Optional[Dict[str, str]]:
        return self.args.k8s_env if self.args else None

    @k8s_env.setter
    def k8s_env(self, k8s_env: Dict[str, str]) -> None:
        if self.args is not None and k8s_env is not None:
            self.args.k8s_env = k8s_env

    @property
    def aws_region(self) -> Optional[str]:
        return self.args.aws_region if self.args else None

    @aws_region.setter
    def aws_region(self, aws_region: str) -> None:
        if self.args is not None and aws_region is not None:
            self.args.aws_region = aws_region

    @property
    def aws_profile(self) -> Optional[str]:
        return self.args.aws_profile if self.args else None

    @aws_profile.setter
    def aws_profile(self, aws_profile: str) -> None:
        if self.args is not None and aws_profile is not None:
            self.args.aws_profile = aws_profile

    @property
    def aws_config_file(self) -> Optional[str]:
        return self.args.aws_config_file if self.args else None

    @aws_config_file.setter
    def aws_config_file(self, aws_config_file: str) -> None:
        if self.args is not None and aws_config_file is not None:
            self.args.aws_config_file = aws_config_file

    @property
    def aws_shared_credentials_file(self) -> Optional[str]:
        return self.args.aws_shared_credentials_file if self.args else None

    @aws_shared_credentials_file.setter
    def aws_shared_credentials_file(self, aws_shared_credentials_file: str) -> None:
        if self.args is not None and aws_shared_credentials_file is not None:
            self.args.aws_shared_credentials_file = aws_shared_credentials_file

    ######################################################
    ## Get App Attributes
    ######################################################

    def get_image_str(self):
        if self.args.image:
            return f"{self.args.image.name}:{self.args.image.tag}"
        elif self.args.image_name and self.args.image_tag:
            return f"{self.args.image_name}:{self.args.image_tag}"
        elif self.args.image_name:
            return f"{self.args.image_name}:latest"
        else:
            return None

    def get_container_name(self) -> str:
        from phidata.utils.common import get_default_container_name

        return self.args.container_name or get_default_container_name(self.name)

    def get_container_port(self) -> int:
        return self.args.container_port

    def get_container_host_port(self) -> int:
        return self.args.container_host_port

    def get_pod_name(self) -> str:
        from phidata.utils.common import get_default_pod_name

        return self.args.pod_name or get_default_pod_name(self.name)

    def get_deploy_name(self) -> str:
        from phidata.utils.common import get_default_deploy_name

        return self.args.deploy_name or get_default_deploy_name(self.name)

    def get_secret_name(self) -> str:
        from phidata.utils.common import get_default_secret_name

        return self.args.secret_name or get_default_secret_name(self.name)

    def get_configmap_name(self) -> str:
        from phidata.utils.common import get_default_configmap_name

        return self.args.configmap_name or get_default_configmap_name(self.name)

    def get_service_name(self) -> str:
        from phidata.utils.common import get_default_service_name

        return self.args.service_name or get_default_service_name(self.name)

    def get_service_port(self) -> int:
        return self.args.service_port

    def get_sa_name(self) -> str:
        from phidata.utils.common import get_default_sa_name

        return self.args.sa_name or get_default_sa_name(self.name)

    def get_cr_name(self) -> str:
        from phidata.utils.common import get_default_cr_name

        return self.args.cr_name or get_default_cr_name(self.name)

    def get_crb_name(self) -> str:
        from phidata.utils.common import get_default_crb_name

        return self.args.crb_name or get_default_crb_name(self.name)

    def get_env_data(self) -> Optional[Dict[str, str]]:
        if self.env_data is None:
            self.env_data = self.read_yaml_file(file_path=self.args.env_file)
        return self.env_data

    def get_secret_data(self) -> Optional[Dict[str, str]]:
        if self.secret_data is None:
            # Read from secrets_file
            self.secret_data = self.read_yaml_file(file_path=self.args.secrets_file)

            # Read from aws_secrets
            if self.args.aws_secrets is not None:
                from phidata.aws.resource.secret.manager import SecretsManager

                aws_secrets: Dict[str, Any] = {}
                if isinstance(self.args.aws_secrets, SecretsManager):
                    _secret_dict = self.args.aws_secrets.get_secrets_as_dict()
                    if _secret_dict is not None and isinstance(_secret_dict, dict):
                        aws_secrets.update(_secret_dict)
                elif isinstance(self.args.aws_secrets, list):
                    for _aws_secret in self.args.aws_secrets:
                        if isinstance(_aws_secret, SecretsManager):
                            _secret_dict = _aws_secret.get_secrets_as_dict()
                            if _secret_dict is not None and isinstance(
                                _secret_dict, dict
                            ):
                                aws_secrets.update(_secret_dict)

                if len(aws_secrets) > 0:
                    if self.secret_data is None:
                        self.secret_data = aws_secrets
                    else:
                        self.secret_data.update(aws_secrets)
        # logger.debug(f"{self.name} secrets: {self.secret_data}")
        return self.secret_data

    def set_container_env(self, container_env: Dict[str, Any]) -> None:
        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file or a secrets backend
        secret_data_from_file = self.get_secret_data()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env
        # this overwrites any existing variables
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(
                {k: str(v) for k, v in self.args.env.items() if v is not None}
            )

    ######################################################
    ## Docker functions
    ######################################################

    def get_docker_container_env(
        self,
        container_paths: ContainerPathContext,
        initial_env: Optional[Dict[str, str]] = None,
    ) -> Dict[str, str]:
        from phidata.constants import (
            PYTHONPATH_ENV_VAR,
            PHIDATA_RUNTIME_ENV_VAR,
            SCRIPTS_DIR_ENV_VAR,
            STORAGE_DIR_ENV_VAR,
            META_DIR_ENV_VAR,
            PRODUCTS_DIR_ENV_VAR,
            NOTEBOOKS_DIR_ENV_VAR,
            WORKFLOWS_DIR_ENV_VAR,
            WORKSPACE_ROOT_ENV_VAR,
            WORKSPACE_CONFIG_DIR_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = initial_env or {}
        container_env.update(
            {
                PHIDATA_RUNTIME_ENV_VAR: "docker",
                SCRIPTS_DIR_ENV_VAR: container_paths.scripts_dir or "",
                STORAGE_DIR_ENV_VAR: container_paths.storage_dir or "",
                META_DIR_ENV_VAR: container_paths.meta_dir or "",
                PRODUCTS_DIR_ENV_VAR: container_paths.products_dir or "",
                NOTEBOOKS_DIR_ENV_VAR: container_paths.notebooks_dir or "",
                WORKFLOWS_DIR_ENV_VAR: container_paths.workflows_dir or "",
                WORKSPACE_ROOT_ENV_VAR: container_paths.workspace_root or "",
                WORKSPACE_CONFIG_DIR_ENV_VAR: container_paths.workspace_config_dir
                or "",
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": container_paths.requirements_file or "",
                "MOUNT_WORKSPACE": str(self.args.mount_workspace),
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            }
        )

        if self.args.set_python_path:
            python_path = self.args.python_path
            if python_path is None:
                python_path = "{}{}".format(
                    container_paths.workspace_root,
                    f":{self.args.add_python_path}"
                    if self.args.add_python_path
                    else "",
                )
            container_env[PYTHONPATH_ENV_VAR] = python_path

        # Set aws env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Set container env using env_file, secrets_file or args.env
        self.set_container_env(container_env=container_env)
        # logger.debug(f"Container env: {container_env}")

        return container_env

    def get_docker_container_volumes(
        self, container_paths: ContainerPathContext
    ) -> Dict[str, dict]:
        from phidata.utils.common import get_default_volume_name

        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = self.args.container_volumes_docker or {}

        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_volume_container_path_str = container_paths.workspace_root
            if (
                self.args.workspace_volume_type is None
                or self.args.workspace_volume_type == WorkspaceVolumeType.HostPath
            ):
                workspace_volume_host_path = (
                    self.args.workspace_volume_host_path
                    or str(self.workspace_root_path)
                )
                logger.debug(f"Mounting: {workspace_volume_host_path}")
                logger.debug(f"\tto: {workspace_volume_container_path_str}")
                container_volumes[workspace_volume_host_path] = {
                    "bind": workspace_volume_container_path_str,
                    "mode": "rw",
                }
            elif self.args.workspace_volume_type == WorkspaceVolumeType.EmptyDir:
                workspace_volume_name = self.args.workspace_volume_name
                if workspace_volume_name is None:
                    workspace_volume_name = get_default_volume_name(
                        container_paths.workspace_name or "ws"
                    )
                logger.debug(f"Mounting: {workspace_volume_name}")
                logger.debug(f"\tto: {workspace_volume_container_path_str}")
                container_volumes[workspace_volume_name] = {
                    "bind": workspace_volume_container_path_str,
                    "mode": "rw",
                }
            else:
                logger.error(f"{self.args.workspace_volume_type.value} not supported")

        return container_volumes

    def get_docker_container_ports(self) -> Dict[str, int]:
        # container_ports is a dictionary which configures the ports to bind
        # inside the container. The key is the port to bind inside the container
        #   either as an integer or a string in the form port/protocol
        # and the value is the corresponding port to open on the host.
        # For example:
        #   {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
        container_ports: Dict[str, int] = self.args.container_ports_docker or {}

        if self.args.open_container_port:
            container_ports[
                str(self.args.container_port)
            ] = self.args.container_host_port

        return container_ports

    def get_container_restart_policy_docker(self) -> Optional[Dict[str, Any]]:
        return self.args.container_restart_policy_docker if self.args else None

    def init_docker_resource_groups(self, docker_build_context: Any) -> None:
        logger.debug(f"@init_docker_resource_groups not defined for {self.name}")

    def get_docker_resource_groups(
        self, docker_build_context: Any
    ) -> Optional[Dict[str, Any]]:
        if self.docker_resource_groups is None:
            self.init_docker_resource_groups(docker_build_context)
        # # Comment out in production
        # if self.docker_resource_groups:
        #     logger.debug("DockerResourceGroups:")
        #     for rg_name, rg in self.docker_resource_groups.items():
        #         logger.debug(
        #             "{}: {}".format(rg_name, rg.json(exclude_none=True, indent=2))
        #         )
        return self.docker_resource_groups

    ######################################################
    ## K8s functions
    ######################################################

    def init_k8s_resource_groups(self, k8s_build_context: Any) -> None:
        logger.debug(f"@init_docker_resource_groups not defined for {self.name}")

    def get_k8s_resource_groups(
        self, k8s_build_context: Any
    ) -> Optional[Dict[str, Any]]:
        if self.k8s_resource_groups is None:
            self.init_k8s_resource_groups(k8s_build_context)
        # # Comment out in production
        # if self.k8s_resource_groups:
        #     logger.debug("K8sResourceGroups:")
        #     for rg_name, rg in self.k8s_resource_groups.items():
        #         logger.debug(
        #             "{}:{}\n{}".format(rg_name, type(rg), rg)
        #         )
        return self.k8s_resource_groups

    ######################################################
    ## AWS functions
    ######################################################

    def get_ecs_container_env(
        self,
        container_paths: ContainerPathContext,
        initial_env: Optional[Dict[str, str]] = None,
    ) -> Dict[str, str]:
        from phidata.constants import (
            PYTHONPATH_ENV_VAR,
            PHIDATA_RUNTIME_ENV_VAR,
            SCRIPTS_DIR_ENV_VAR,
            STORAGE_DIR_ENV_VAR,
            META_DIR_ENV_VAR,
            PRODUCTS_DIR_ENV_VAR,
            NOTEBOOKS_DIR_ENV_VAR,
            WORKFLOWS_DIR_ENV_VAR,
            WORKSPACE_ROOT_ENV_VAR,
            WORKSPACE_CONFIG_DIR_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = initial_env or {}
        container_env.update(
            {
                PHIDATA_RUNTIME_ENV_VAR: "ecs",
                SCRIPTS_DIR_ENV_VAR: container_paths.scripts_dir or "",
                STORAGE_DIR_ENV_VAR: container_paths.storage_dir or "",
                META_DIR_ENV_VAR: container_paths.meta_dir or "",
                PRODUCTS_DIR_ENV_VAR: container_paths.products_dir or "",
                NOTEBOOKS_DIR_ENV_VAR: container_paths.notebooks_dir or "",
                WORKFLOWS_DIR_ENV_VAR: container_paths.workflows_dir or "",
                WORKSPACE_ROOT_ENV_VAR: container_paths.workspace_root or "",
                WORKSPACE_CONFIG_DIR_ENV_VAR: container_paths.workspace_config_dir
                or "",
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": container_paths.requirements_file or "",
                "MOUNT_WORKSPACE": str(self.args.mount_workspace),
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            }
        )

        if self.args.set_python_path:
            python_path = self.args.python_path
            if python_path is None:
                python_path = "{}{}".format(
                    container_paths.workspace_root or "",
                    f":{self.args.add_python_path}"
                    if self.args.add_python_path
                    else "",
                )
            container_env[PYTHONPATH_ENV_VAR] = python_path

        # Set aws env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Set container env using env_file, secrets_file or args.env
        self.set_container_env(container_env=container_env)
        # logger.debug(f"Container env: {container_env}")

        return container_env

    def init_aws_resource_groups(self, aws_build_context: Any) -> None:
        logger.debug(f"@init_aws_resource_groups not defined for {self.name}")

    def get_aws_resource_groups(
        self, aws_build_context: Any
    ) -> Optional[Dict[str, Any]]:
        if self.aws_resource_groups is None:
            self.init_aws_resource_groups(aws_build_context)
        # Comment out in production
        # if self.aws_resource_groups:
        #     logger.debug("qAwsResourceGroups:")
        #     for rg_name, rg in self.aws_resource_groups.items():
        #         try:
        #             logger.debug("{}\n{}".format(rg_name, rg.json(indent=2)))
        #         except Exception:
        #             pass
        return self.aws_resource_groups

    ######################################################
    ## Helpers
    ######################################################

    def get_container_paths(self, add_ws_name_to_ws_root: bool = True) -> Optional[Any]:
        if self.workspace_root_path is None:
            return None

        workspace_name = self.workspace_root_path.stem
        if workspace_name is None:
            return None

        workspace_volume_container_path: str = self.args.workspace_volume_container_path
        if workspace_volume_container_path is None:
            return None

        from phidata.types.context import ContainerPathContext

        workspace_root_container_path = (
            f"{workspace_volume_container_path}/{workspace_name}"
            if add_ws_name_to_ws_root
            else workspace_volume_container_path
        )
        container_paths = ContainerPathContext(
            workspace_name=workspace_name,
            workspace_root=workspace_root_container_path,
            workspace_parent=workspace_volume_container_path,
        )

        if self.args.scripts_dir is not None:
            container_paths.scripts_dir = (
                f"{workspace_root_container_path}/{self.args.scripts_dir}"
            )

        if self.args.storage_dir is not None:
            container_paths.storage_dir = (
                f"{workspace_root_container_path}/{self.args.storage_dir}"
            )

        if self.args.meta_dir is not None:
            container_paths.meta_dir = (
                f"{workspace_root_container_path}/{self.args.meta_dir}"
            )

        if self.args.products_dir is not None:
            container_paths.products_dir = (
                f"{workspace_root_container_path}/{self.args.products_dir}"
            )

        if self.args.notebooks_dir is not None:
            container_paths.notebooks_dir = (
                f"{workspace_root_container_path}/{self.args.notebooks_dir}"
            )

        if self.args.workflows_dir is not None:
            container_paths.workflows_dir = (
                f"{workspace_root_container_path}/{self.args.workflows_dir}"
            )

        if self.args.workspace_config_dir is not None:
            container_paths.workspace_config_dir = (
                f"{workspace_root_container_path}/{self.args.workspace_config_dir}"
            )

        if self.args.requirements_file is not None:
            container_paths.requirements_file = (
                f"{workspace_root_container_path}/{self.args.requirements_file}"
            )

        return container_paths

    def read_yaml_file(self, file_path: Optional[Path]) -> Optional[Dict[str, Any]]:
        if file_path is not None and file_path.exists() and file_path.is_file():
            import yaml

            # logger.debug(f"Reading {file_path}")
            data_from_file = yaml.safe_load(file_path.read_text())
            if data_from_file is not None and isinstance(data_from_file, dict):
                return data_from_file
            else:
                logger.error(f"Invalid file: {file_path}")
        return None

    def set_aws_env_vars(self, env_dict: Dict[str, str]) -> None:
        """Set AWS environment variables."""
        from phidata.constants import (
            AWS_REGION_ENV_VAR,
            AWS_DEFAULT_REGION_ENV_VAR,
            # AWS_PROFILE_ENV_VAR,
            # AWS_CONFIG_FILE_ENV_VAR,
            # AWS_SHARED_CREDENTIALS_FILE_ENV_VAR,
        )

        if self.aws_region is not None:
            env_dict[AWS_REGION_ENV_VAR] = self.aws_region
            env_dict[AWS_DEFAULT_REGION_ENV_VAR] = self.aws_region
        # if self.aws_profile is not None:
        #     env_dict[AWS_PROFILE_ENV_VAR] = self.aws_profile
        # if self.aws_config_file is not None:
        #     env_dict[AWS_CONFIG_FILE_ENV_VAR] = self.aws_config_file
        # if self.aws_shared_credentials_file is not None:
        #     env_dict[
        #         AWS_SHARED_CREDENTIALS_FILE_ENV_VAR
        #     ] = self.aws_shared_credentials_file



from phidata.app.jupyter.jupyter import Jupyter, JupyterArgs
from phidata.app.jupyter.jupyter_lab import (
    JupyterLab,
    JupyterLabArgs,
    ImagePullPolicy,
    ServiceType,
    RestartPolicy,
)
from phidata.app.jupyter.jupyter_hub import JupyterHub, JupyterHubArgs



from typing import Tuple

from phidata.utils.log import logger

try:
    import duckdb
except ImportError:
    logger.warning("DuckDB not available.")
    raise


def load_s3_path_to_table(
    duckdb_connection: duckdb.DuckDBPyConnection, load_s3_path: str
) -> Tuple[str, str]:
    """
    Load a file from S3 into duckdb

    Args:
        duckdb_connection (duckdb.DuckDBPyConnection): duckdb connection
        load_s3_path (str): S3 path to load

    Returns:
        str: Table name
        str: SQL statement used to load the file
    """
    import os

    # Get the file name from the s3 path
    file_name = load_s3_path.split("/")[-1]
    # Get the file name without extension from the s3 path
    table_name, extension = os.path.splitext(file_name)
    # If the table_name isn't a valid SQL identifier, we'll need to use something else
    table_name = (
        table_name.replace("-", "_")
        .replace(".", "_")
        .replace(" ", "_")
        .replace("/", "_")
    )

    create_statement = (
        f"CREATE OR REPLACE TABLE '{table_name}' AS SELECT * FROM '{load_s3_path}';"
    )
    duckdb_connection.sql(create_statement)

    return table_name, create_statement



from pathlib import Path
from typing import Optional, Dict, Any, List, Union

from phidata.app.aws_app import AwsApp, AwsAppArgs
from phidata.app.base_app import WorkspaceVolumeType
from phidata.app.docker_app import DockerApp, DockerAppArgs
from phidata.app.k8s_app import (
    K8sApp,
    K8sAppArgs,
    ImagePullPolicy,
    RestartPolicy,
    ServiceType,
)
from phidata.utils.log import logger


class StreamlitAppArgs(AwsAppArgs, DockerAppArgs, K8sAppArgs):
    pass


class StreamlitApp(AwsApp, DockerApp, K8sApp):
    def __init__(
        self,
        name: str = "streamlit",
        version: str = "1",
        enabled: bool = True,
        # -*- Image Configuration,
        # Image can be provided as a DockerImage object or as image_name:image_tag
        image: Optional[Any] = None,
        image_name: str = "phidata/streamlit",
        image_tag: str = "1.23",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = "streamlit hello",
        # -*- Debug Mode
        debug_mode: bool = False,
        # -*- Python Configuration,
        # Install python dependencies using a requirements.txt file,
        install_requirements: bool = False,
        # Path to the requirements.txt file relative to the workspace_root,
        requirements_file: str = "requirements.txt",
        # Set the PYTHONPATH env var,
        set_python_path: bool = False,
        # Manually provide the PYTHONPATH,
        python_path: Optional[str] = None,
        # Add paths to the PYTHONPATH env var,
        # If python_path is provided, this value is ignored,
        add_python_paths: Optional[List[str]] = None,
        # -*- Container Environment,
        # Add env variables to container,
        env: Optional[Dict[str, Any]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Add secret variables to container,
        secrets: Optional[Dict[str, Any]] = None,
        # Read secret variables from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Read secret variables from AWS Secrets,
        aws_secrets: Optional[Any] = None,
        # -*- Container Ports,
        # Open a container port if open_container_port=True,
        open_container_port: bool = True,
        # Port number on the container,
        container_port: int = 9095,
        # Port name (only used by the K8sContainer),
        container_port_name: str = "http",
        # Host port to map to the container port,
        container_host_port: int = 9095,
        # -*- Workspace Volume,
        # Mount the workspace directory on the container,
        mount_workspace: bool = False,
        workspace_volume_name: Optional[str] = None,
        workspace_volume_type: Optional[WorkspaceVolumeType] = None,
        # Path to mount the workspace volume inside the container,
        workspace_dir_container_path: str = "/usr/local/app",
        # Add the workspace name to the container path,
        add_workspace_name_to_container_path: bool = False,
        # -*- If workspace_volume_type=WorkspaceVolumeType.HostPath,
        # Mount workspace_dir to workspace_dir_container_path,
        # If None, use the workspace_root,
        workspace_dir: Optional[str] = None,
        # -*- Container Configuration,
        container_name: Optional[str] = None,
        # Run container in the background and return a Container object.,
        container_detach: bool = True,
        # Enable auto-removal of the container on daemon side when the container’s process exits.,
        container_auto_remove: bool = True,
        # Remove the container when it has finished running. Default: True.,
        container_remove: bool = True,
        # Username or UID to run commands as inside the container.,
        container_user: Optional[Union[str, int]] = None,
        # Keep STDIN open even if not attached.,
        container_stdin_open: bool = True,
        # Return logs from STDOUT when container_detach=False.,
        container_stdout: Optional[bool] = True,
        # Return logs from STDERR when container_detach=False.,
        container_stderr: Optional[bool] = True,
        container_tty: bool = True,
        # Specify a test to perform to check that the container is healthy.,
        container_healthcheck: Optional[Dict[str, Any]] = None,
        # Optional hostname for the container.,
        container_hostname: Optional[str] = None,
        # Platform in the format os[/arch[/variant]].,
        container_platform: Optional[str] = None,
        # Path to the working directory.,
        container_working_dir: Optional[str] = None,
        # Add labels to the container,
        container_labels: Optional[Dict[str, str]] = None,
        # Restart the container when it exits. Configured as a dictionary with keys:,
        # Name: One of on-failure, or always.,
        # MaximumRetryCount: Number of times to restart the container on failure.,
        # For example: {"Name": "on-failure", "MaximumRetryCount": 5},
        container_restart_policy: Optional[Dict[str, Any]] = None,
        # Add volumes to DockerContainer,
        # container_volumes is a dictionary which adds the volumes to mount,
        # inside the container. The key is either the host path or a volume name,,
        # and the value is a dictionary with 2 keys:,
        #   bind - The path to mount the volume inside the container,
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.,
        # For example:,
        # {,
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},,
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'},
        # },
        container_volumes: Optional[Dict[str, dict]] = None,
        # Add ports to DockerContainer,
        # The keys of the dictionary are the ports to bind inside the container,,
        # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.,
        # The values of the dictionary are the corresponding ports to open on the host, which can be either:,
        #   - The port number, as an integer.,
        #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.,
        #   - None, to assign a random host port. For example, {'2222/tcp': None}.,
        #   - A tuple of (address, port) if you want to specify the host interface.,
        #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.,
        #   - A list of integers, if you want to bind multiple host ports to a single container port.,
        #       For example, {'1111/tcp': [1234, 4567]}.,
        container_ports: Optional[Dict[str, Any]] = None,
        # -*- Pod Configuration,
        pod_name: Optional[str] = None,
        pod_annotations: Optional[Dict[str, str]] = None,
        pod_node_selector: Optional[Dict[str, str]] = None,
        # -*- Secret Configuration,
        secret_name: Optional[str] = None,
        # -*- Configmap Configuration,
        configmap_name: Optional[str] = None,
        # -*- Deployment Configuration,
        replicas: int = 1,
        deploy_name: Optional[str] = None,
        # Type: ImagePullPolicy,
        image_pull_policy: Optional[Any] = None,
        # Type: RestartPolicy,
        deploy_restart_policy: Optional[Any] = None,
        deploy_labels: Optional[Dict[str, Any]] = None,
        termination_grace_period_seconds: Optional[int] = None,
        # Key to spread the pods across a topology,
        topology_spread_key: Optional[str] = None,
        # The degree to which pods may be unevenly distributed,
        topology_spread_max_skew: Optional[int] = None,
        # How to deal with a pod if it doesn't satisfy the spread constraint.,
        topology_spread_when_unsatisfiable: Optional[str] = None,
        # -*- Service Configuration,
        create_service: bool = False,
        service_name: Optional[str] = None,
        # Type: ServiceType,
        service_type: Optional[Any] = None,
        # The port exposed by the service.,
        service_port: int = 9095,
        # The node_port exposed by the service if service_type = ServiceType.NODE_PORT,
        service_node_port: Optional[int] = None,
        # The target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        service_target_port: Optional[Union[str, int]] = None,
        # Extra ports exposed by the webserver service. Type: List[CreatePort],
        service_ports: Optional[List[Any]] = None,
        service_labels: Optional[Dict[str, Any]] = None,
        service_annotations: Optional[Dict[str, str]] = None,
        # If ServiceType == ServiceType.LoadBalancer,
        service_health_check_node_port: Optional[int] = None,
        service_internal_traffic_policy: Optional[str] = None,
        service_load_balancer_class: Optional[str] = None,
        service_load_balancer_ip: Optional[str] = None,
        service_load_balancer_source_ranges: Optional[List[str]] = None,
        service_allocate_load_balancer_node_ports: Optional[bool] = None,
        # -*- Ingress Configuration,
        create_ingress: bool = False,
        ingress_name: Optional[str] = None,
        ingress_annotations: Optional[Dict[str, str]] = None,
        # -*- RBAC Configuration,
        use_rbac: bool = False,
        # Create a Namespace with name ns_name & default values,
        ns_name: Optional[str] = None,
        # or Provide the full Namespace definition,
        # Type: CreateNamespace,
        namespace: Optional[Any] = None,
        # Create a ServiceAccount with name sa_name & default values,
        sa_name: Optional[str] = None,
        # or Provide the full ServiceAccount definition,
        # Type: CreateServiceAccount,
        service_account: Optional[Any] = None,
        # Create a ClusterRole with name cr_name & default values,
        cr_name: Optional[str] = None,
        # or Provide the full ClusterRole definition,
        # Type: CreateClusterRole,
        cluster_role: Optional[Any] = None,
        # Create a ClusterRoleBinding with name crb_name & default values,
        crb_name: Optional[str] = None,
        # or Provide the full ClusterRoleBinding definition,
        # Type: CreateClusterRoleBinding,
        cluster_role_binding: Optional[Any] = None,
        # -*- AWS Configuration,
        aws_subnets: Optional[List[str]] = None,
        aws_security_groups: Optional[List[Any]] = None,
        # -*- ECS Configuration,
        ecs_cluster: Optional[Any] = None,
        ecs_launch_type: str = "FARGATE",
        ecs_task_cpu: str = "1024",
        ecs_task_memory: str = "2048",
        ecs_service_count: int = 1,
        assign_public_ip: bool = True,
        ecs_enable_exec: bool = True,
        # -*- LoadBalancer Configuration,
        load_balancer: Optional[Any] = None,
        listener: Optional[Any] = None,
        # Create a load balancer if load_balancer is None,
        create_load_balancer: bool = False,
        # HTTP or HTTPS,
        load_balancer_protocol: str = "HTTP",
        load_balancer_security_groups: Optional[List[Any]] = None,
        # Default 80 for HTTP and 443 for HTTPS,
        load_balancer_port: Optional[int] = None,
        load_balancer_certificate: Optional[Any] = None,
        load_balancer_certificate_arn: Optional[str] = None,
        # -*- TargetGroup Configuration,
        target_group: Optional[Any] = None,
        # HTTP or HTTPS,
        target_group_protocol: str = "HTTP",
        # Default 80 for HTTP and 443 for HTTPS,
        target_group_port: Optional[int] = None,
        target_group_type: str = "ip",
        health_check_protocol: Optional[str] = None,
        health_check_port: Optional[str] = None,
        health_check_enabled: Optional[bool] = None,
        health_check_path: Optional[str] = None,
        health_check_interval_seconds: Optional[int] = None,
        health_check_timeout_seconds: Optional[int] = None,
        healthy_threshold_count: Optional[int] = None,
        unhealthy_threshold_count: Optional[int] = None,
        #  -*- Resource Control,
        skip_create: bool = False,
        skip_read: bool = False,
        skip_update: bool = False,
        recreate_on_update: bool = False,
        skip_delete: bool = False,
        wait_for_creation: bool = True,
        wait_for_update: bool = True,
        wait_for_deletion: bool = True,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 50,
        #  -*- Save Resources to output directory,
        # If True, save the resources to files,
        save_output: bool = False,
        # The resource directory for the output files,
        resource_dir: Optional[str] = None,
        # Skip creation if resource with the same name is active,
        use_cache: bool = True,
        #  -*- Other args,
        print_env_on_load: bool = False,
        # Extra kwargs used to capture additional args,
        **extra_kwargs,
    ):
        super().__init__()

        try:
            self.args: StreamlitAppArgs = StreamlitAppArgs(
                name=name,
                version=version,
                enabled=enabled,
                image=image,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                debug_mode=debug_mode,
                install_requirements=install_requirements,
                requirements_file=requirements_file,
                set_python_path=set_python_path,
                python_path=python_path,
                add_python_paths=add_python_paths,
                env=env,
                env_file=env_file,
                secrets=secrets,
                secrets_file=secrets_file,
                aws_secrets=aws_secrets,
                open_container_port=open_container_port,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_volume_type=workspace_volume_type,
                workspace_dir_container_path=workspace_dir_container_path,
                add_workspace_name_to_container_path=add_workspace_name_to_container_path,
                workspace_dir=workspace_dir,
                container_name=container_name,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                container_user=container_user,
                container_stdin_open=container_stdin_open,
                container_stdout=container_stdout,
                container_stderr=container_stderr,
                container_tty=container_tty,
                container_healthcheck=container_healthcheck,
                container_hostname=container_hostname,
                container_platform=container_platform,
                container_working_dir=container_working_dir,
                container_labels=container_labels,
                container_restart_policy=container_restart_policy,
                container_volumes=container_volumes,
                container_ports=container_ports,
                pod_name=pod_name,
                pod_annotations=pod_annotations,
                pod_node_selector=pod_node_selector,
                secret_name=secret_name,
                configmap_name=configmap_name,
                replicas=replicas,
                deploy_name=deploy_name,
                image_pull_policy=image_pull_policy,
                deploy_restart_policy=deploy_restart_policy,
                deploy_labels=deploy_labels,
                termination_grace_period_seconds=termination_grace_period_seconds,
                topology_spread_key=topology_spread_key,
                topology_spread_max_skew=topology_spread_max_skew,
                topology_spread_when_unsatisfiable=topology_spread_when_unsatisfiable,
                create_service=create_service,
                service_name=service_name,
                service_type=service_type,
                service_port=service_port,
                service_node_port=service_node_port,
                service_target_port=service_target_port,
                service_ports=service_ports,
                service_labels=service_labels,
                service_annotations=service_annotations,
                service_health_check_node_port=service_health_check_node_port,
                service_internal_traffic_policy=service_internal_traffic_policy,
                service_load_balancer_class=service_load_balancer_class,
                service_load_balancer_ip=service_load_balancer_ip,
                service_load_balancer_source_ranges=service_load_balancer_source_ranges,
                service_allocate_load_balancer_node_ports=service_allocate_load_balancer_node_ports,
                create_ingress=create_ingress,
                ingress_name=ingress_name,
                ingress_annotations=ingress_annotations,
                use_rbac=use_rbac,
                ns_name=ns_name,
                namespace=namespace,
                sa_name=sa_name,
                service_account=service_account,
                cr_name=cr_name,
                cluster_role=cluster_role,
                crb_name=crb_name,
                cluster_role_binding=cluster_role_binding,
                aws_subnets=aws_subnets,
                aws_security_groups=aws_security_groups,
                ecs_cluster=ecs_cluster,
                ecs_launch_type=ecs_launch_type,
                ecs_task_cpu=ecs_task_cpu,
                ecs_task_memory=ecs_task_memory,
                ecs_service_count=ecs_service_count,
                assign_public_ip=assign_public_ip,
                ecs_enable_exec=ecs_enable_exec,
                load_balancer=load_balancer,
                listener=listener,
                create_load_balancer=create_load_balancer,
                load_balancer_protocol=load_balancer_protocol,
                load_balancer_security_groups=load_balancer_security_groups,
                load_balancer_port=load_balancer_port,
                load_balancer_certificate=load_balancer_certificate,
                load_balancer_certificate_arn=load_balancer_certificate_arn,
                target_group=target_group,
                target_group_protocol=target_group_protocol,
                target_group_port=target_group_port,
                target_group_type=target_group_type,
                health_check_protocol=health_check_protocol,
                health_check_port=health_check_port,
                health_check_enabled=health_check_enabled,
                health_check_path=health_check_path,
                health_check_interval_seconds=health_check_interval_seconds,
                health_check_timeout_seconds=health_check_timeout_seconds,
                healthy_threshold_count=healthy_threshold_count,
                unhealthy_threshold_count=unhealthy_threshold_count,
                skip_create=skip_create,
                skip_read=skip_read,
                skip_update=skip_update,
                recreate_on_update=recreate_on_update,
                skip_delete=skip_delete,
                wait_for_creation=wait_for_creation,
                wait_for_update=wait_for_update,
                wait_for_deletion=wait_for_deletion,
                waiter_delay=waiter_delay,
                waiter_max_attempts=waiter_max_attempts,
                save_output=save_output,
                resource_dir=resource_dir,
                use_cache=use_cache,
                print_env_on_load=print_env_on_load,
                **extra_kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid: {e}")
            raise



from phidata.app.server.server_base import (
    ServerBase,
    ServerBaseArgs,
    ServiceType,
    RestartPolicy,
    ImagePullPolicy,
)
from phidata.app.server.api_server import ApiServer, ApiServerArgs
